# Authors:
#   Jason Gerard DeRose <jderose@redhat.com>
#
# wehjit: A Python web-widget library
# Copyright (C) 2009  Red Hat
#
# This file is part of `wehjit`.
#
# `wehjit` is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# `wehjit` is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# `wehjit`.  If not, see <http://www.gnu.org/licenses/>.

"""
Collect pages together in an `Application`.
"""

import sys
import os
from os import path
import shutil
import time
from cStringIO import StringIO

from assetslib import Assets

from freezable import Freezable, freeze
from styles import Styles
from util import not_found
from collection import Collection, builtins
import base
import errors


class Application(Freezable):
    def __init__(self, url='/', assets=None, widgets=builtins, prod=True, **kw):
        """
        Initialize.

        :param url: The URL this application will be mounted at, e.g.
            ``'/my-app/'``; defaults to ``'/'``

        :param assets: An ``assetslib.Assets`` instance providing assets for
            this application; by default an ``Assets`` instance is created, but
            this created instance is not suitable for production deployment as
            the assets will be stored in a different random temporary directory
            per process

        :param widgets: A `Collection` instance providing the widget plugins
            from which new widget instances are created; defaults to
            `collection.builtins`

        :param prod: If ``True``, put the `Application` instance in "production"
            mode, which means that all FireBug console calls are stripped from
            JavaScript assets upon rendering; defaults to ``True``

        :param kw: Additional keyword arguments to pass to new `base.Page`
            instances
        """
        if url:
            self.url = url.rstrip('/') + '/'
        else:
            self.url = ''
        if assets is None:
            if self.url.startswith('/'):
                u = '/_/'
            else:
                u = '_/'
            self.assets = Assets(u)
        else:
            self.assets = assets
        self.widgets = widgets
        self.prod = prod
        self.__kw = kw
        self.styles = Styles()
        self.__plugins = set()
        self.__pages = list()
        self.__ids = dict()

    def __repr__(self):
        return '%s(url=%r)' % (self.__class__.__name__, self.url)

    def __iter__(self):
        """
        Iterate in ascending order through the IDs of the pages in App.

        Because a `base.Page` instance can have it's ID changed after it was
        created via `Application.new()`, the ID-to-Page map isn't populated till
        `Application.finalize()` is called, before which this method will always
        return an empty generator.

        For example:

        >>> app = Application()
        >>> p1 = app.new('Page', id='foo:bar')
        >>> p2 = app.new('Page', id='bar:baz')
        >>> list(app)
        []
        >>> app.finalize()
        >>> list(app)
        ['bar:baz', 'foo:bar']
        """
        for id in sorted(self.__ids):
            yield id

    def __contains__(self, id):
        """
        Return ``True`` if App. contains a `base.Page` instance with ID *id*.

        Because a `base.Page` instance can have it's ID changed after it was
        created via `Application.new()`, the ID-to-Page map isn't populated till
        `Application.finalize()` is called, before which this method will always
        return ``False``.

        For example:

        >>> app = Application()
        >>> 'foo:bar' in app
        False
        >>> p = app.new('Page', id='foo:bar')
        >>> 'foo:bar' in app
        False
        >>> app.finalize()
        >>> 'foo:bar' in app
        True


        :param id: The ID of the page in question, e.g. ``'my:test:page'``
        """
        return id in self.__ids

    def __getitem__(self, id):
        """
        Return the `base.Page` instance with ID *id*.

        Because a `base.Page` instance can have it's ID changed after it was
        created via `Application.new()`, the ID-to-Page map isn't populated till
        `Application.finalize()` is called, before which this method will always
        raise a ``KeyError``.

        For example:

        >>> app = Application()
        >>> app.new('Page', id='foo:bar')
        Page(id='foo:bar')
        >>> app['foo:bar']
        Traceback (most recent call last):
          ...
        KeyError: 'foo:bar'
        >>> app.finalize()
        >>> app['foo:bar']
        Page(id='foo:bar')

        After `Application.finalize()` has been called, this method will only
        raise a ``KeyError`` if no page exists with an ID equal to *id*.  For
        example:

        >>> app['bar:baz']
        Traceback (most recent call last):
          ...
        KeyError: 'bar:baz'


        :param id: The ID of the page to retrieve, e.g. ``'my:test:page'``
        """
        return self.__ids[id]

    def __len__(self):
        return len(self.__pages)

    def print_tree(self, fileobj=sys.stdout, level=0):
        indent = ' ' * (level * 4)
        fileobj.write(indent + repr(self) + '\n')
        for page in self.__pages:
            page.print_tree(fileobj, level + 1)

    def plugins(self):
        """
        Iterate through the plugins that have been used at least once.

        The `Application` can only track which plugins are used if you create
        widget instances via `Application.new()`.  If you've created them
        another way, they wont show up here.

        Also note that an `Application` may use fewer plugins than available in
        its widgets `Collection`.
        """
        for plugin in sorted(self.__plugins, key=lambda p: p.__name__):
            yield plugin

    def pages(self):
        """
        Iterate through all `wehjit.Page` instances.

        Note that the `Application` can only track which which pages you've
        created if you create them via `Application.new()`.  If you've created
        them another way, they wont show up here.
        """
        for page in self.__pages:
            yield page

    def new(self, klass, **kw):
        """
        Create an instance of the widget plugin named *klass*, passing it *kw*.

        In most ways creating a widgets directly from a `Collection`, like this:

        >>> widgets = Collection('my_widgets')
        >>> widgets.register_builtins()
        >>> foo = widgets.Reset(name='foo')

        Is equivalent to creating a widget using `Application.new()`, like this:

        >>> app = Application(widgets=widgets)
        >>> bar = app.new('Submit', name='bar')

        The difference is that in the 1st case, the `Application` can't track
        what plugins have been used:

        >>> list(app.plugins())  # Wont contain `wehjit.plugins.form.Reset`:
        [<class 'wehjit.plugins.form.Submit'>]

        Nor can the `Application` provide the new widget a reference to itself
        via the *app* kwarg:

        >>> foo.app is None
        True
        >>> bar.app is app
        True

        Additionally, all `base.Page` instances are tracked if created with
        `Application.new()`.  For example:

        >>> p1 = widgets.Page(id='eg:one')
        >>> p2 = app.new('Page', id='eg:two')
        >>> list(app.pages())  # Wont contain p1
        [Page(id='eg:two')]
        >>> app.finalize()
        >>> list(app)  # Wont contain 'eg:one'
        ['eg:two']


        :param klass: The name of the widget plugin to instantiate, e.g.
            ``'Page'`` or ``'Button'``
        :param kw: Additional keyword arguments to supply to widget constructor
        """
        kw['app'] = self
        plugin = self.widgets[klass]
        self.__plugins.add(plugin)
        if issubclass(plugin, base.Page):
            kw['baseurl'] = self.url
            kw.update(self.__kw)
            inst = plugin(**kw)
            self.__pages.append(inst)
        else:
            inst = plugin(**kw)
        return inst

    def finalize(self):
        """
        Prepare application for serving or rendering to static files.

        This method will initialize the ID-to-page map, call
        `base.Page.finalize()` on all `base.Page` instances, and lastly put this
        `Application` into a thread-safe state.
        """
        for page in self.__pages:
            if page.id in self.__ids:
                raise errors.PageConflict(
                    new=page, orig=self.__ids[page.id], app=self
                )
            self.__ids[page.id] = page
        for page in self.__pages:
            page.finalize()
        freeze(self)

    def render_static(self, out='static'):
        class Reporter(object):
            def __init__(self, task, total, tab=2):
                self.__start = time.time()
                self.task = task
                self.total = total
                self.i = 0
                self.tab = ' ' * tab
                self._print_h1(self.task)

            @property
            def elapsed(self):
                return time.time() - self.__start

            @property
            def minsec(self):
                sec = self.elapsed
                return (sec / 60, sec % 60)

            def _print_h1(self, msg):
                msg += '.'
                print '\n' + msg
                print '=' * len(msg)

            def step(self, msg, *args):
                msg %= args
                self.i += 1
                print '\n%d/%d %s...' % (self.i, self.total, msg)

            def end(self, msg='Finished'):
                msg += (' in %d:%02d' % self.minsec)
                self._print_h1(msg)

            def info(self, msg, *args):
                msg %= args
                print self.tab + msg

        reporter = Reporter('Rendering app to static files', len(self) + 1)
        reporter.info('target=%r', path.abspath(out))
        for key in sorted(self.__kw):
            reporter.info('%s=%r', key, self.__kw[key])

        if path.lexists(out):
            raise StandardError('out: output path exists: %r' % out)
        os.makedirs(out)
        assert path.isdir(out)

        # Copy assets over to static dst
        reporter.step('Copying assets')
        if self.assets.url.startswith('/'):
            prefix = 'abs'
        else:
            prefix = 'rel'
        parts = [prefix]
        parts.extend(self.assets.url.split('/'))
        parts = filter(None, parts)
        dst = path.join(out, *parts)
        reporter.info('url=%r', self.assets.url)
        reporter.info('dir=%r', dst)
        shutil.copytree(self.assets.dir, dst, symlinks=True)

        # Render pages
        for page in self.pages():
            page.render_static(out, reporter)

        reporter.end('Rendered app to static pages')
        reporter.info('result=%r', path.abspath(out))
        for key in sorted(self.__kw):
            reporter.info('%s=%r', key, self.__kw[key])

    def render_assets(self):
        """
        Render all assets.

        This method renders the "magic" CSS asset and the "magic" JavaScript
        asset based on the total plugins that are used (instantiated at least
        once) in this `Application`.  Plugins can only be tracked if widget
        instances are created via `Application.new()`.

        This method will also register/render non-magic assets by calling the
        ``register_assets()`` classmethod on any plugins that have defined it.

        If you're running under a multi-process server, don't call this method
        when each process starts as you'll simply be overwriting identical
        assets rendered when earlier processes started.  Ideally your should
        render your assets at application install time (perhaps with a hook into
        your distributions package system).
        """
        for plugin in self.plugins():
            self._update_style(plugin)
            if callable(plugin.register_assets):
                plugin.register_assets(self.assets)
        self._register_asset('wehjit.css', self._render_css)
        self._register_asset('wehjit.js', self._render_js)

    def _update_style(self, plugin):
        if plugin.style_vars:
            self.styles.update_vars(**dict(plugin.style_vars))
        if plugin.style_global:
            for (selector, definition) in plugin.style_global:
                self.styles.update_global(selector, definition)
        if plugin.style:
            for (selector, definition) in plugin.style:
                self.styles.update(plugin.__name__, selector, definition)

    def _register_asset(self, key, method):
        buf = StringIO()
        method(buf)
        data = buf.getvalue()
        if data:
            self.assets.register(data=data, key=key)

    def _render_css(self, fileobj):
        self.styles.render_global(fileobj)
        for plugin in self.plugins():
            self.styles.render(fileobj, plugin.__tag__, plugin.__name__)

    def _render_js(self, fileobj):
        rendered = set()
        page = []
        other = []
        for p in self.plugins():
            if issubclass(p, base.Page):
                page.append(p)
            else:
                other.append(p)
        for p in page + other:
            if p.javascript and p.javascript not in rendered:
                rendered.add(p.javascript)
                fileobj.write('/*  ----  %s  ----  */\n' % p.__name__)
                if self.prod:
                    self._strip_firebug_console(fileobj, p.javascript)
                else:
                    fileobj.write(p.javascript.encode('utf-8'))

    def _strip_firebug_console(self, fileobj, data):
        for line in data.splitlines(True):
            s = line.strip()
            if s.startswith('console.'):
                assert s.endswith(';')
                continue
            fileobj.write(line.encode('utf-8'))

    def __call__(self, environ, start_response):
        id = environ['PATH_INFO'].strip('/').replace('/', ':')
        if id in self:
            page = self[id]
            environ['SCRIPT_NAME'] += environ['PATH_INFO']
            environ['PATH_INFO'] = ''
            return page(environ, start_response)
        return not_found(environ, start_response)
