# Authors:
#   Jason Gerard DeRose <jderose@redhat.com>
#
# wehjit: A Python web-widget library
# Copyright (C) 2009  Red Hat
#
# This file is part of `wehjit`.
#
# `wehjit` is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# `wehjit` is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# `wehjit`.  If not, see <http://www.gnu.org/licenses/>.

"""
FreeIPA.
"""

from wehjit.compat import json
from wehjit import base, builtins
from wehjit.state import *


class Grid(base.Widget):
    cols = Static('cols', json=True, default=tuple())
    rows = Static('rows', json=True, default=tuple())
    searchable = Static('searchable', default=False)

    message = Dynamic('message', doc='Message text to display')
    error = Dynamic('error', default=True, doc='If True, message is an error')

    xml = """
    <div
        xmlns:py="http://genshi.edgewall.org/"
        class="${klass}"
        id="${id}"
    >
    <div
        class="${['message', 'message error'][bool(error)]}"
        py:content="message"
    />
    <table>

    <thead>

    <tr>
    <th
        py:for="col in cols"
        title="Click to sort"
        class="sortable"
        name="${col.name}"
    >
    <span class="arrow">&darr;</span>
    <span class="text" py:content="col.label" />
    &nbsp;
    </th>
    <th class="space" />
    </tr>

    <tr py:if="searchable">
    <td py:for="col in cols">
    <input type="text" name="${col.name}" class="${col.css_classes}" />
    </td>
    </tr>

    </thead>

    <tbody>
    <tr>
    <td
        py:for="col in cols"
        name="${col.name}"
        class="${col.css_classes}"
    />
    <td class="space" />
    </tr>
    </tbody>

    </table>
    </div>
    """

    style_vars = (
        ('colorbg_grid_th', '#777'),
        ('colorbg_grid_th_hover', '#977'),
        ('colorbg_grid_tr', '#eee'),
        ('colorbg_grid_tr_hover', '#ddd'),
        ('colorbg_grid_tr_selected', '#ddf'),
    )

    style = (
        ('', (
            ('position', 'fixed'),
            ('top', '2em'),
            ('bottom', '0'),
            ('left', '0'),
            ('right', '0'),
        )),

        ('div.message', (
            ('text-align', 'center'),
            ('height', '%(height_bar)s'),
            ('line-height', '%(height_bar)s'),
            ('font-weight', 'bold'),
        )),

        ('div.message.error', (
            ('color', '%(color_error)s !important'),
        )),

        ('table', (
            ('position', 'absolute'),
            ('display', 'block'),
            ('top', '2em'),
            ('bottom', '0'),
            ('left', '0'),
            ('right', '0'),
            ('overflow-x', 'auto'),
            ('overflow-y', 'hidden'),
            ('empty-cells', 'show'),
            ('border-collapse', 'separate'),
            ('border-spacing', '2px'),
            ('cursor', 'default'),
            ('white-space', 'nowrap'),
            ('font-size', '%(font_size_mono)s'),
            ('font-family', 'monospace'),
        )),

        ('thead', (
            ('background-color', '%(colorbg_grid_th)s'),
        )),

        ('thead input', (
            ('border', '0'),
        )),

        ('tbody', (
            ('overflow-x', 'hidden'),
            ('overflow-y', 'auto'),
        )),

        ('tr', (
            ('line-height', '2em'),
        )),

        ('tbody tr.selected', (
            ('background-color', '%(colorbg_grid_tr_selected)s !important'),
        )),

        ('tbody tr', (
            ('background-color', '%(colorbg_grid_tr)s'),
        )),

        ('tbody tr:hover', (
            ('background-color', '%(colorbg_grid_tr_hover)s'),
        )),

        ('th', (
            ('color', '#fff'),
            ('padding', '0 0.5em'),
            ('text-align', 'left'),
        )),

        ('th.sort', (
            ('color', 'yellow'),
        )),

        ('span.arrow', (
            ('font-size', '12pt'),
            ('visibility', 'hidden'),
        )),

        ('th.sort span.arrow', (
            ('visibility', 'visible'),
        )),

        ('th.sortable:hover', (
            ('background-color', '%(colorbg_grid_th_hover)s'),
        )),

        ('td', (
            ('padding', '0 0.5em'),
        )),

        ('th.space', (
            ('width', '100%%'),
        )),

    )


    js_class = 'Grid'
    javascript = u"""
    Wehjit.bases.Grid = new Class({
        Extends: Wehjit.bases.Widget,

        post_init: function() {
            if (this.tag == 'table') {
                this.table = this.el;
            }
            else {
                this.table = this.el.getElement('table');
            }
            this.thead = this.table.getElement('thead');
            this.tbody = this.table.getElement('tbody');
            this.tr = this.tbody.getElement('tr').dispose();
            console.assert(this.table);
            console.assert(this.thead);
            console.assert(this.tbody);
            console.assert(this.tr);
            this.message = this.el.getElement('div.message');
            console.assert(this.message);
            this.message.set('tween', {duration: 1500, link: 'cancel'});

            window.addEvent('resize', this.on_resize.bind(this));
            this.tbody.addEvent('click', this.on_tbody_click.bind(this));
            this.tbody.addEvent('dblclick', this.on_tbody_dblclick.bind(this));

            this.sorted_by = null;
            this.reversed = false;

            this.sortables = new Hash();
            this.thead.getElements('th.sortable').each(function(th) {
                th.connect('click', this, 'on_th_click');
                this.sortables.set(th.getAttribute('name'), th);
            }, this);
        },

        post_init2: function() {
            if ($type(this.data.rows) == 'array') {
                this.redraw();
            }
        },

        set_summary: function(text, highlight) {
            if (! text) {
                this.message.textContent = '';
                return;
            }
            this.message.removeClass('error');
            this.message.textContent = text;
            if (highlight) {
                this.message.highlight();
                //this.message.tween('color', '#0f0', '#000');
            }
        },

        set_error: function(text) {
            if (! text) {
                this.message.textContent = '';
                return;
            }
            this.message.addClass('error');
            this.message.textContent = text;
        },

        set_rows: function(rows) {
            this.data.rows = rows;
            this.redraw();
        },

        redraw: function() {
            if ($type(this.data.rows) != 'array') {
                this.data.rows = [];
            }
            console.time('redraw');
            this.tbody.set_height('auto');
            this.select(null);
            while (this.tbody.rows.length) {
                this.tbody.deleteRow(-1);
            }
            for (var i=0; i < this.data.rows.length; i ++) {
                var row = this.data.rows[i];
                var tr = this.tr.cloneNode(true);
                row['__tr__'] = tr;
                tr.__row__ = row;
                for (var j=0; j < tr.cells.length - 1; j++) {
                    var td = tr.cells[j];
                    var value = row[td.getAttribute('name')];
                    td.set_text(value);
                }
                this.tbody.appendChild(tr);
            }
            this.resize();
            console.timeEnd('redraw');
        },

        on_resize: function() {
            this.resize();
        },

        resize: function() {
            console.log('scrollHeight = %d', this.tbody.scrollHeight);
            this.tbody.set_height(Math.min(
                this.tbody.scrollHeight,
                this.table.clientHeight - this.thead.clientHeight
            ));
        },

        on_tbody_click: function(event) {
            if (! event.target.tagName == 'td') {
                return;
            }
            var tr = event.target.parentNode;
            console.assert(tr.tagName == 'tr');
            if (tr.__row__ != this.selected) {
                this.select(tr.__row__);
            }
        },

        select: function(row) {
            if (this.selected) {
                this.selected.__tr__.removeClass('selected');
            }
            if ($type(row) == 'number') {
                var row = this.data.rows[row];
            }
            this.selected = row;
            if (row && row.__tr__) {
                row.__tr__.addClass('selected');
            }
            this.fireEvent('select', row);
        },

        scrollto: function() {
            if (this.selected && this.selected.__tr__) {
                this.selected.__tr__.scrollIntoView();
            }
        },

        activate: function(row) {
            if (row != this.selected) {
                this.select(row);
            }
            this.fireEvent('activate', row);
        },

        on_tbody_dblclick: function(event) {
            if (event.target.tagName != 'td') {
                return;
            }
            var tr = event.target.parentNode;
            console.assert(tr.tagName == 'tr');
            console.assert(tr.__row__ == this.selected);
            this.activate(this.selected);
        },

        sort: function() {
            /*
            Sort this.data.rows based on current this.sorted_by, this.reversed.
            */
            if (! this.sorted_by) {
                return;
            }
            var name = this.sorted_by;
            if (this.reversed) {
                var f = function(a, b) {
                    if (a[name] > b[name]) {
                        return -1;
                    }
                    if (a[name] < b[name]) {
                        return 1;
                    }
                    return 0;
                };
            }
            else {
                var f = function(a, b) {
                    if (a[name] < b[name]) {
                        return -1;
                    }
                    if (a[name] > b[name]) {
                        return 1;
                    }
                    return 0;
                };
            }
            this.data.rows.sort(f);
        },

        sync_header_state: function() {
            /*
            Make the table headers reflect the current sort state.
            */
            this.sortables.each(function(other) {
                other.removeClass('sort');
            });
            var th = this.sortables.get(this.sorted_by);
            if (! th) {
                return;
            }
            th.addClass('sort');
            var arrow = this.reversed ? '\u2191' : '\u2193' ;
            th.getElement('span.arrow').textContent = arrow;
        },

        no_sort: function() {
            this.sorted_by = null;
            this.sync_header_state();
        },

        on_th_click: function(th, event) {
            if ($type(this.data.rows) != 'array') {
                return;
            }
            console.time('on_th_click');
            var name = th.getAttribute('name');
            if (name == this.sorted_by) {
                this.reversed = ! this.reversed;
            }
            this.sorted_by = name;
            this.sort(name);
            this.sync_header_state();
            while (this.tbody.rows.length) {
                this.tbody.deleteRow(-1);
            }
            for (var i=0; i < this.data.rows.length; i ++) {
                var row = this.data.rows[i];
                this.tbody.appendChild(row['__tr__']);
            }
            this.select(null);
            console.timeEnd('on_th_click');
        },

    });
    """

builtins.register(Grid)


class CRUDS(base.Widget):
    key = Static('key', defult='primary_key', json=True)
    display_cols = Static('display_cols', json=True, default=tuple())

    label_create = Static('label_create', default='Create')
    label_retrieve = Static('label_retrieve', default='Retrieve')
    label_update = Static('label_update', default='Update')
    label_delete = Static('label_delete', default='Delete')
    label_search = Static('label_search', default='Search:')

    jsonrpc_url = Static('jsonrpc_url', json=True)
    method_create = Static('method_create', default='create', json=True)
    method_retrieve = Static('method_retrieve', default='retrieve', json=True)
    method_update = Static('method_update', default='update', json=True)
    method_delete = Static('method_delete', default='delete', json=True)
    method_search = Static('method_search', default='search', json=True)

    autoload = Static('autoload', default=False, json=True,
        doc='If True, run search at page load',
    )

    xml = """
    <div
        xmlns:py="http://genshi.edgewall.org/"
        class="${css_classes}"
        id="${id}"
    >
    <button name="create" py:content="label_create" />
    <button name="retrieve" py:content="label_retrieve" />
    <button name="update" py:content="label_update" />
    <button name="delete" py:content="label_delete" />
    <label>${label_search}<input type="text" /></label>
    </div>
    """

    js_class = 'CRUDS'

    javascript = """
    Wehjit.bases.CRUDS = new Class({
        Extends: Wehjit.bases.Widget,

        post_init: function() {
            this.grid = Wehjit.get('grid');
            console.assert(this.grid);
            this.grid.addEvent('select', this.on_select.bind(this));
            this.grid.addEvent('activate', this.on_retrieve.bind(this));

            this.dialogs = Wehjit.get('dialogs');
            console.assert(this.dialogs);

            this.el.getElements('button').each(function(el) {
                var name = el.getAttribute('name');
                console.assert(name);
                this['button_' + name] = el;
                el.addEvent('click', this['on_' + name].bind(this));
            }, this);
            this.search = this.el.getElement('input');
            this.search.addEvent('change', this.on_search_change.bind(this));
            this.rpc_id = 0;
            this.search.focus();
        },

        post_init2: function() {
            ['create', 'retrieve', 'update', 'delete', 'result'].each(
                function(name) {
                    this[name] = this.dialogs.get(name);
                    this[name].addEvent('close', this.on_close.bind(this));
            }, this);

            this.create.addEvent('create', this.do_create.bind(this));
            this.retrieve.addEvent('update', this.on_update.bind(this));
            this.retrieve.addEvent('delete', this.on_delete.bind(this));
            this.update.addEvent('update', this.do_update.bind(this));
            this.delete.addEvent('delete', this.do_delete.bind(this));

            this.form_create = this.create.get_children()[0];
            console.assert(this.form_create.klass == 'Form');

            this.form_update = this.update.get_children()[0];
            console.assert(this.form_update.klass == 'Form');

            if (this.grid.data.rows.length == 0 && this.data.autoload) {
                this.on_search_change();
            }
        },

        get_value: function() {
            if (! (this.row && this.data.key)) {
                return null;
            }
            var value = this.row[this.data.key];
            if ($type(value) == 'array') {
                return value.join('');
            }
            return value;
        },

        send: function(action, args, options) {
            if (this.request) {
                this.request.cancel();
            }
            if ($type(this.data.jsonrpc_url) != 'string') {
                console.error('%s is missing jsonrpc_url', this.id);
                return;
            }
            var method = this.data['method_' + action];
            if ($type(method) != 'string') {
                console.error('%s has no method for %s', this.id, action);
                return;
            }
            this.action = action;
            var data = {
                method: method,
                params: [args, options],
                id: this.rpc_id++,
            };
            var options = {
                url: this.data.jsonrpc_url,
                data: JSON.encode(data),
                urlEncoded: false,
                headers: {'Content-Type': 'application/json'},
            };
            this.request = new Request.JSON(options);
            this.request.addEvent('success', this.on_success.bind(this));
            this.request.addEvent('failure', this.on_failure.bind(this));
            this.request.send();
        },

        on_success: function(response, responseText) {
            this.request = null;
            if ($type(response.error) == 'object') {
                // Assume this is a JSON-RPC 2.0 response
                var error = response.error.message;
            }
            else {
                var error = response.error;
            }
            var d = this.dialogs.current;
            if (error) {
                if (d) {
                    d.set_error(error);
                }
                else {
                    this.grid.set_error(error);
                }
                return;
            }
            if (this.action == 'search') {
                this.grid.set_summary(response.result.summary);
                this.grid.set_rows(response.result.result);
                return;
            }
            this.grid.set_summary('');
            this.dialogs.close();
            this.update_grid(response);
            // this.result.set_summary(response.result.summary);
            // this.dialogs.run('result', response.result.result);
        },

        on_failure: function(xhr) {
            var error = 'Request failed to ' + this.data.jsonrpc_url;
            console.error(error);
            if (this.dialogs.current) {
                this.dialogs.current.set_error(error);
            }
            else {
                this.grid.set_error(error);
            }
        },

        update_grid: function(response) {
            var entry = response.result.result;
            switch (this.action) {
                case 'create':
                    this.form_create.reset();
                    this.grid.data.rows.splice(0, 0, entry);
                    this.grid.sort();
                    this.grid.redraw();
                    this.grid.select(entry);
                    this.grid.scrollto();
                    break;
                case 'update':
                    var i = this.grid.data.rows.indexOf(this.row);
                    this.grid.data.rows[i] = entry;
                    this.grid.sort();
                    this.grid.redraw();
                    this.grid.select(entry);
                    this.grid.scrollto();
                    break;
                case 'delete':
                    var i = this.grid.data.rows.indexOf(this.row);
                    console.log('deleting %s', i);
                    this.grid.data.rows.splice(i, 1);
                    this.grid.redraw();
                    break;
            }
            this.grid.set_summary(response.result.summary, true);
        },

        on_close: function(dialog, params) {
            this.search.focus();
        },

        on_select: function(row) {
            this.row = row;
            var disabled = (row) ? false : true;
            ['retrieve', 'update', 'delete'].each(function(name) {
                this['button_' + name].disabled = disabled;
            }, this);
        },

        on_activate: function(row) {
            this.dialogs.run('retrieve', row)
        },

        on_create: function() {
            this.create.set_summary('');
            this.dialogs.run('create');
        },

        do_create: function() {
            this.send('create', [], this.form_create.get_data());
        },

        on_retrieve: function() {
            this.retrieve.format_title({value: this.get_value()});
            this.dialogs.run('retrieve', this.row);
        },

        on_update: function() {
            this.update.set_summary('');
            this.update.format_title({value: this.get_value()});
            this.form_update.set_data(this.row);
            this.dialogs.run('update', this.row);
        },

        do_update: function() {
            this.send('update', [this.get_value()], this.form_update.get_data());
        },

        on_delete: function() {
            this.create.set_summary('');
            this.delete.format_title({value: this.get_value()});
            this.dialogs.run('delete', this.row);
        },

        do_delete: function() {
            console.log('do_delete');
            this.send('delete', [this.get_value()], {});
        },

        on_search_change: function() {
            console.log('search change');
            this.send('search', [this.search.value], {});
        },

    });
    """

builtins.register(CRUDS)



class Display(base.Widget):
    xml = """
    <table
        xmlns:py="http://genshi.edgewall.org/"
        class="${css_classes}"
        id="${id}"
    />
    """

    style = (
        ('', (
            ('empty-cells', 'show'),
            ('border-collapse', 'collapse'),
            ('margin', '1em 2em'),
        )),

        ('tr.error', (
            ('background-color', '#ddd'),
        )),


        ('th', (
            ('text-align', 'right'),
            ('padding', '0.25em 0.5em'),
            ('vertical-align', 'top'),
            ('line-height', '%(height_bar)s'),
        )),

        ('td', (
            ('padding', '0.25em 0.5em'),
            ('vertical-align', 'top'),
            ('text-align', 'left'),
            ('line-height', '%(height_bar)s'),
        )),

    )

    js_class = 'Display'

    javascript = """
    Wehjit.bases.Display = new Class({
        Extends: Wehjit.bases.Widget,

        post_init: function() {
            var parent = this.get_parent();
            console.assert(parent);
            parent.addEvent('run', this.on_run.bind(this));
            this.cruds = Wehjit.get('cruds');
            this.cols = this.cruds.data.display_cols;
            console.assert(this.cols);
            if (this.cols.length == 0) {
                this.cols = Wehjit.data.grid.cols;
            }
        },

        on_run: function(dialog, row) {
            console.log('Display.on_run(%s, %s)', dialog, row);
            this.el.empty();
            if ($type(row) != 'object') {
                return;
            }
            this.cols.each(function(col) {
                var tr = new Element('tr');
                var th = new Element('th');
                th.set_text(col.label);
                tr.appendChild(th);
                this.el.appendChild(tr);
                var td = new Element('td');
                td.set_text(row[col.name]);
                tr.appendChild(td);
            }, this);
        },

    });
    """

builtins.register(Display)
